#!/usr/bin/env python
"""
Module to manipulate git repositories and their hooks.
"""

# We'll need the schema module to play with the database.
from schema import db
import schema
repos = schema.get_table('repositories')
repo_acls = schema.get_table('repository_acls')
users = schema.get_table('users') # repository ACLs refer to users

repo_root = "/opt/git"
base_hook_dir = 'default_hooks' # name of directory from which we grab hook code

# We'll also need some SQLAlchemy functionality to write queries.
from sqlalchemy.sql import select
from sqlalchemy.sql.expression import and_, or_, join, outerjoin, subquery

# We do a lot of path manipulation...
import os

def format_name(name):
    # append .git extension if missing; it seems to be a convention
    if len(name) < 4 or name[-4:] != ".git":
        name += ".git"

    # make sure the path derives from the root
    # TODO: we need canonicalization here to prevent traversal attacks
    path = os.path.join(repo_root, name)

    return (name, path)

def get_repo(name):
    s = select([repos]).where(repos.c.repository_name == name)
    result = db.execute(s)
    return result.fetchone()

def get_repo_by_id(id):
    s = select([repos]).where(repos.c.repository_id == id)
    result = db.execute(s)
    return result.fetchone()

def delete_repo_by_id(id):
    repo = get_repo_by_id(id)
    if not repo:
        raise KeyError("Invalid repository ID")
    (name, path) = format_name(repo['repository_name'])
    import shutil
    shutil.rmtree(path)
    # TODO: Shouldn't these two be in the same transaction?
    #       How do we implement transactions in SQLAlchemy?
    s = repos.delete().where(repos.c.repository_id == id)
    result = db.execute(s)
    s = repo_acls.delete().where(repo_acls.c.repository_id == id)
    result = db.execute(s)
    # TODO: return a value in this function?

def get_repos_for_user(user_id):
    columns = repos.c + [repo_acls.c.is_owner, repo_acls.c.can_read,
        repo_acls.c.can_write, repo_acls.c.can_rewind,
        repo_acls.c.can_create_tag, repo_acls.c.can_modify_tag]
    s = select(columns).where(
        (repo_acls.c.repository_id == repos.c.repository_id)
        & (repo_acls.c.user_id == user_id)
        & (repo_acls.c.can_read | repo_acls.c.is_owner))
    return db.execute(s).fetchall()

# TODO: Everything related to repository ACLs will not scale with user count.
#       Eventually, we need to figure out some sort of pagination here.
def get_repo_acls(repo_id, user_id=None):
    """
    Get a list of the permissions each user has on this repo.

    If user_id is provided, the returned list will only contain that user's
    ACLs.
    """
    # branch if filtering by user ID
    user_clause = True
    if user_id:
        user_clause = users.c.user_id == user_id

    
    sq = (
        select([repo_acls])
        .where(repo_acls.c.repository_id == repo_id)
        .alias()
    )
    columns = [c for c in users.c]
    columns += [c for c in sq.c if c not in [sq.c.user_id]]
    s = (
        users.outerjoin(sq, (users.c.user_id == sq.c.user_id))
        .select(user_clause)
        .with_only_columns(columns)
    )
    result = db.execute(s)
    return result.fetchall()

def update_repo_acls(repo_id, user_id, **kwargs):
    # check if we need to do an update
    # TODO: This behavior is ANYTHING except transaction-safe...
    s = select([repo_acls]).where(
        (repo_acls.c.repository_id == repo_id)
        & (repo_acls.c.user_id == user_id))
    acl = db.execute(s).fetchone()
    if not acl:
        s = repo_acls.insert().values(
            repository_id=repo_id,
            user_id=user_id,
            **kwargs)
    else:
        s = repo_acls.update().values(**kwargs).where(
            (repo_acls.c.repository_id == repo_id)
            & (repo_acls.c.user_id == user_id)
        )
    db.execute(s)

def update_hooks(name):
    (name, path) = format_name(name)
    gitzebo_dir = os.path.dirname(os.path.abspath(__file__))
    base_hook_path = os.path.join(gitzebo_dir, base_hook_dir)
    from glob import glob
    from shutil import copyfile, copymode
    from tempfile import NamedTemporaryFile
    from textwrap import dedent
    for source_path in glob(os.path.join(base_hook_path, '*')):
        basename = os.path.basename(source_path)

        # Perform inline edits to insure that our environment is
        # reproduced when git invokes the hooks.  These hacks insure that
        # a virtualenv installation of gitzebo won't choke when hooks
        # execute -- which helps, as virtualenv is the suggested deployment
        # path.
        temp_file = NamedTemporaryFile(mode='w')
        with open(source_path, 'r') as source_file:
            for line in source_file:
                # TODO: Doesn't the VIRTUAL_ENV variable get passed by the
                #       wrapper?  If so, we likely don't need any of this
                #       hacky silliness anymore...
                if line.strip() == '# INSERT ENVIRONMENT HACKS HERE':
                    print "Adding environment hacks"
                    virtualenv = os.getenv('VIRTUAL_ENV', None)
                    if virtualenv:
                        activate_path = os.path.join(
                            virtualenv, 'bin', 'activate_this.py')
                        environment_lines = """\
                            # Activate virtualenv
                            activator = '{0}'
                            execfile(activator, dict(__file__=activator))
                            """.format(activate_path)
                        # Prevent wild indenting from triple-quotes
                        environment_lines = dedent(environment_lines)
                        temp_file.write(environment_lines)
                        print "Setting virtualenv to " + virtualenv
                    else:
                        print "No special environment detected;",
                        print "not writing environment hacks."
                else:
                    temp_file.write(line)
        temp_file.flush()

        # Copy our modified hook over.  Moving might screw up the
        # temp file's removal when it shifts out of scope.
        target_path = os.path.join(path, 'hooks', basename)
        print source_path
        print " ->", temp_file.name, "(for editing atomically)"
        print " ->", target_path
        # Copy the contents of the temp file...
        copyfile(temp_file.name, target_path)
        # ... but the mode of the source file; otherwise the hooks don't tend
        # to be executable.
        copymode(source_path, target_path)

# Creating a new repository
def create_repo(name, owner_id):
    (name, path) = format_name(name)

    # TODO: a synchronization problem exists here when:
    #    - the repository already exists, but doesn't appear in the DB
    #    - the repository is missing, but appears in the DB
    # If one of these conditions exists, what do we do?
    #
    # For now:
    #    - if it exists in FS but not DB, populate DB
    #    - if it exists in DB but not FS, initialize repo in FS
    #
    # Basically, we naively chug on through and initialize all uninitialized
    # pieces without throwing exceptions.  I'm not sure this is the best
    # policy, as desyncs here may indicate further consistency issues, and
    # ignoring them may hamper debugging.

    # error out if the repository already exists in both FS and DB
    repo = get_repo(name)
    if os.path.isdir(path) and repo:
        raise KeyError("Repository {0} already exists.".format(name))

    # if no FS directory exists, initialize the repo /w subprocess'ed git
    if not os.path.isdir(path):
        # TODO: can we use a native Python library here instead of subprocess?
        # TODO: do we need to insure git is installed properly somehow?
        import subprocess
        cmd = ['git', 'init', '--bare', path]
        proc = subprocess.Popen(
            args=cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()
        if proc.returncode != 0:
            raise Exception("git init {0} failed: {1}".format(path, stderr))

    # insert repository row
    s = repos.insert().values(
        repository_name=name
    )
    result = db.execute(s)
    repo_id = result.inserted_primary_key[0]

    # insert default owner ACL
    # TODO: Shouldn't this be in the same transaction as above?!
    #       Theoretically a failure here could be unrecoverable.
    s = repo_acls.insert().values(
        user_id=owner_id,
        repository_id=repo_id,
        is_owner=True,
        can_write=True,
        can_rewind=True,
        can_read=True,
        can_create_tag=True,
        can_modify_tag=True,
    )
    result = db.execute(s)

    # make sure all hooks are up-to-date
    update_hooks(name)

    return repo_id




#!/usr/bin/env python
from flask import Flask, render_template
app = Flask(__name__)

from schema import db

import flask
from flask import request, redirect, abort

#######################
# Authentication code #
#######################
# Pulled from http://flask.pocoo.org/snippets/31/
# TODO: POSSIBLY ready to refactor out
from functools import wraps
from flask import request, Response
from users import verify_user

def authenticate():
    """
    Requests authentication credentials from user.
    """
    return Response('You must log in to access this functionality.',
        401, {'WWW-Authenticate': 'Basic realm="Gitzebo"'})

@app.before_request
def auth_helper():
    request.user = None
    auth = request.authorization
    if not auth:
        return
    user = verify_user(auth.username, auth.password)
    if not user:
        return
    request.user = user
    return

def requires_auth(f):
    """
    Decorates a Flask call to require authentication.
    """
    @wraps(f)
    def decorated(*args, **kwargs):
        if not request.user:
            return authenticate()
        return f(*args, **kwargs)
    return decorated

########################
# CORE WEB APPLICATION #
########################
import users, repos, keys

@app.route("/")
def home():
    return redirect(url_for('list_repos'))

#####################
# USER MANIPULATION #
#####################

@app.route("/users")
@requires_auth
def list_users():
    # return list of users
    userlist = users.get_users()
    return render_template('users.html', userlist=userlist)

@app.route("/users/<int:id>")
@requires_auth
def user_details(id):
    user = users.get_user_by_id(id)
    return render_template('user_details.html', user=user)

@app.route("/users/add", methods=['POST'])
@requires_auth
def add_user():
    if not request.user['can_create_users']:
        raise Exception("Permission denied; user cannot modify users.")
    for key in ['user_name', 'user_password', 'commit_name', 'commit_email']:
        if key not in request.form or len(request.form[key]) < 1:
            raise Exception("User creation form requires field '{0}'"
                .format(key))
    user_name = request.form['user_name']
    password = request.form['user_password']
    commit_name = request.form['commit_name']
    commit_email = request.form['commit_email']
    user_id = users.create_user(
        user_name,
        password=password,
        commit_name=commit_name,
        commit_email=commit_email)
    # TODO: redirect to user details?
    return redirect(url_for('list_users'))

@app.route("/users/delete/<int:id>")
@requires_auth
def delete_user(id):
    if not request.user['can_create_users']:
        raise Exception("Permission denied; user cannot modify users.")
    users.delete_user(id)
    return redirect(url_for('list_users'))

####################
# KEY MANIPULATION #
####################

@app.route("/keys")
@requires_auth
def list_keys():
    keylist = keys.get_keys(request.user['user_id'])
    return render_template('keys.html', keylist=keylist)

@app.route("/keys/add", methods=['POST'])
@requires_auth
def add_key():
    # create_key(user_id, key, name):
    keys.create_key(
        user_id=request.user['user_id'],
        key=request.form['key'].strip(),
        name=request.form['key_name'].strip(),
    )
    return redirect(url_for('list_keys'))

@app.route("/keys/delete/<int:id>")
@requires_auth
def delete_key(id):
    key = keys.get_key_by_id(id)
    if not key:
        raise Exception("Key does not exist!")
    if key['user_id'] != request.user['user_id']:
        raise Exception("Permission denied: you don't own this key.")
    keys.delete_key(id)
    return redirect(url_for('list_keys'))


###########################
# REPOSITORY MANIPULATION #
###########################

@app.route("/repos")
@requires_auth
def list_repos():
    repolist = repos.get_repos_for_user(request.user['user_id'])
    return render_template('repos.html', repolist=repolist)

def assert_repo_permission(repo_id, permission):
    user_acls = repos.get_repo_acls(
        repo_id=repo_id,
        user_id=request.user['user_id'])

    user_acls = user_acls[0]

    if not user_acls[permission]:
        raise Exception("Permission denied; user {0} lacks {1} permissions"
            " for the repository with ID {2}".format(
                request.user['user_name'],
                permission,
                repo_id))

def render_repo_details(id, template):
    repo = repos.get_repo_by_id(id)
    repo_acls = repos.get_repo_acls(repo_id=repo['repository_id'])
    user_id = request.user['user_id']
    user_acls = [a for a in repo_acls if a['user_id'] == user_id][0]

    # get our hostname for the git clone example
    import urlparse
    parsed_url = urlparse.urlparse(request.url)

    return render_template(template,
        repo=repo,
        repo_acls=repo_acls,
        hostname=parsed_url.hostname,
        user_acls=user_acls)

@app.route("/repos/details/<int:id>")
@requires_auth
def repo_details(id):
    assert_repo_permission(id, 'can_read')
    return render_repo_details(id, 'repo_details.html')

@app.route("/repos/edit/<int:id>", methods=['GET', 'POST'])
@requires_auth
def edit_repo(id):
    assert_repo_permission(id, 'is_owner')

    if request.method == 'GET':
        return render_repo_details(id, 'repo_edit.html')

    # commit changes
    permissions = ['is_owner', 'can_read', 'can_write', 'can_rewind',
        'can_create_tag', 'can_modify_tag']
    for user in users.get_users():
        user_permissions = {}
        if 'exists_' + str(user['user_id']) not in request.form:
            continue
        for permission in permissions:
            name = '{0}_{1}'.format(permission, user['user_id'])
            value = name in request.form
            user_permissions[permission] = value
        repos.update_repo_acls(
            repo_id=id,
            user_id=user['user_id'],
            **user_permissions
        )

    return redirect(url_for('edit_repo', id=id))

# TODO: Pretty much all of our administration tasks are vulnerable to CSRF.
@app.route("/repos/delete/<int:id>")
@requires_auth
def delete_repo(id):
    repo = repos.get_repo_by_id(id)
    repo_acls = repos.get_repo_acls(
        repo_id=repo['repository_id'],
        user_id=request.user['user_id'])
    if not repo_acls or len(repo_acls) < 1 or not repo_acls[0]['is_owner']:
        raise Exception("Permission denied; user doesn't own this repository.")
    repos.delete_repo_by_id(id)
    return redirect(url_for('list_repos'))

@app.route("/repos/add", methods=['POST'])
@requires_auth
def add_repo():
    if not request.user['can_create_repositories']:
        raise Exception("Permission denied; user cannot create repositories.")
    name = request.form['repository_name']
    # TODO: externalize minimum repo name length
    if not name or len(name) <= 3:
        raise Exception("Repository name is too short.")
    id = repos.create_repo(name=name, owner_id=request.user['user_id'])
    return redirect(url_for('repo_details', id=id))

####################
# TEMPLATE HELPERS #
####################
@app.template_global()
def url_for(target, *args, **kwargs):
    return flask.url_for(target, *args, **kwargs)




#!/usr/bin/env python
# We'll need the schema module to play with the database.
from schema import db
import schema
keys = schema.get_table('keys')
users = schema.get_table('users') # used when generating authorized_keys

# We'll also need some SQLAlchemy functionality to write queries.
from sqlalchemy.sql import select

# TODO: We assume RSA keys at the moment.

def validate_key(key):
    """
    Validate that a given string looks like a proper public key.
    """
    # Validate that the string parses
    import base64, binascii
    try:
        base64.decodestring(key)
    except binascii.Error:
        return False

    # Validate the Base64 alphabet described in RFC 3548
    valid_chars = [chr(ord('A') + i) for i in range(26)]
    valid_chars += [chr(ord('a') + i) for i in range(26)]
    valid_chars += [chr(ord('0') + i) for i in range(10)]
    valid_chars += ['+', '/', '=']
    for char in key:
        if char not in valid_chars:
            return False

    # If nothing has bounced, we're as okay as we can imagine
    return True

def create_key(user_id, key, name):
    # validate data
    if not validate_key(key):
        raise ValueError("Invalid public key: '" + key + "'")

    # validate uniqueness
    if get_key_by_key(key):
        raise Exception("Key already exists in database.")

    # insert key
    stmt = keys.insert().values(
        user_id=user_id,
        name=name,
        public_key=key)
    result = db.execute(stmt)

    # insure the filesystem reflects changes and return
    regenerate_authorized_keys()
    return result.inserted_primary_key

def delete_key(key_id):
    # delete it!
    s = keys.delete().where(keys.c.key_id == key_id)
    result = db.execute(s)
    if result.rowcount != 1:
        raise Exception("Deleted {0} rows, expected to delete one.".format(
            result.rowcount))

    # insure the filesystem reflects changes and return
    regenerate_authorized_keys()
    return result.rowcount

def get_key_by_key(key):
    """
    Look up a key by the actual key (reverse lookup).

    Used to enforce uniqueness of keys, so that we don't generate an
    invalid authorized_keys file.
    """
    s = select([keys]).where(keys.c.public_key == key)
    result = db.execute(s)
    return result.fetchone()

def get_key_by_id(key_id):
    s = select([keys]).where(keys.c.key_id == key_id)
    result = db.execute(s)
    return result.fetchone()

def get_keys(user_id):
    s = select([keys]).where(keys.c.user_id == user_id)
    result = db.execute(s)
    return result.fetchall()

def regenerate_authorized_keys():
    # os will be used for path manipulation and environment reading
    import os

    # do as much validaton as possible
    home = os.getenv('HOME', None)
    if home is None:
        raise ValueError("HOME environment variable should be set!")

    # perform SQL query
    s = select([keys, users]).where(keys.c.user_id == users.c.user_id)
    result = db.execute(s)

    # attempt to locate our SSH wrapper
    from distutils.spawn import find_executable
    wrapper_path = find_executable('gitzebo-ssh-wrapper')

    # write resultset to file
    target_dir = os.path.join(home, '.ssh')
    target_path = os.path.join(target_dir, 'authorized_keys')
    #gitzebo_dir = os.path.dirname(os.path.abspath(__file__))
    #wrapper_path = os.path.join(gitzebo_dir, 'gitzebo-ssh-wrapper')
    # TODO: eventually allow for DSA keys, too?
    venv = ''
    venv_dir = os.getenv('VIRTUAL_ENV', None)
    if venv_dir:
        venv = ' VIRTUAL_ENV=' + venv_dir
    with open(target_path, 'w') as f:
        for row in db.execute(s):
            f.write('command="{wrapper} USERNAME={username}{venv}",'
                'no-port-forwarding,'
                'no-X11-forwarding,'
                'no-agent-forwarding,'
                'no-pty'
                ' ssh-rsa {key}\n'.format(
                    wrapper=wrapper_path,
                    username=row['user_name'],
                    key=row['public_key'],
                    venv=venv, # virtualenv spec, if applicable
                )
            )

    # move temporary file into place and return
    # TODO: We should be doing atomic writes here!
    #       Write to a temp file, and then overwrite-move it to the target path.




#!/usr/bin/env python
import sqlalchemy, os

# DB path is currently rooted in our module directory by default
# TODO: database configuration should be externalized
db_path = os.path.join(os.path.dirname(__file__), 'gitzebo.db')
db = sqlalchemy.create_engine("sqlite:///{0}".format(db_path))
latest_version=1

from sqlalchemy import (Table, Column, MetaData,
    ForeignKey, ColumnDefault,
    Integer, String, Boolean, Text)

def get_metadata(version=latest_version, db=db):
    metadata = MetaData(db)

    schema_version = Table('_schema_version', metadata,
        Column('version', Integer),
    )

    users = Table('users', metadata,
        Column('user_id', Integer, primary_key=True),
        Column('user_name', String, unique=True),
        Column('pass_hash', String),
        Column('pass_salt', String),
        Column('commit_name', String),
        Column('commit_email', String),
        Column('can_create_users', Boolean, ColumnDefault(False)),
        Column('can_create_repositories', Boolean, ColumnDefault(False)),
    )

    keys = Table('keys', metadata,
        Column('key_id', Integer, primary_key=True),
        Column('user_id', Integer, ForeignKey('users.user_id')),
        Column('name', String),
        Column('public_key', Text),
        # TODO: public_key should be unique, but TEXT columns can't be (I think)
        #       verify me maybe?
    )

    repos = Table('repositories', metadata,
        Column('repository_id', Integer, primary_key=True),
        Column('repository_name', String, unique=True),
    )

    repo_acls = Table('repository_acls', metadata,
        Column('user_id', Integer, ForeignKey('users.user_id')),
        Column('repository_id', Integer,
            ForeignKey('repositories.repository_id')),
        Column('is_owner', Boolean, ColumnDefault(False)),
        Column('can_write', Boolean, ColumnDefault(False)),
        Column('can_rewind', Boolean, ColumnDefault(False)),
        Column('can_read', Boolean, ColumnDefault(False)),
        Column('can_create_tag', Boolean, ColumnDefault(False)),
        Column('can_modify_tag', Boolean, ColumnDefault(False)),
    )

    return metadata

def get_table(table, db=db):
    metadata = get_metadata(db=db)
    return metadata.tables[table]

def create_schema(version=latest_version, db=db):
    metadata = get_metadata(version=version, db=db)

    # Insure that our schema versioning table doesn't exist--
    # which also insures that we're not trying to provision over
    # an existing database, as we have an explicit upgrade
    # workflow.
    if metadata.tables['_schema_version'].exists():
        raise Exception("Database already provisioned")

    # Check to see if any tables already exist, which would mean
    # we're attempting to provision a database that contains some
    # other application's data.  If it contained our application's
    # data, the _schema_version table would exist!
    for table in metadata.sorted_tables:
        if table.exists():
            raise Exception("Table already exists: " + table.name)

    # Create all of our tables
    metadata.create_all(db)

    # Populate schema version
    version_table = metadata.tables['_schema_version']
    # TODO: do we need to check for success somehow?
    db.execute(version_table.insert().values(version=version))

    # Create the default admin user
    from users import create_user
    id = create_user('admin', password='admin', commit_email='', can_create_users=True, can_create_repositories=True)
    id = create_user('jgilik', password='test', commit_name='John Gilik', commit_email='john@jgilik.com', can_create_users=True, can_create_repositories=True)
    id = create_user('test', password='test', commit_name='Tester McDonald', commit_email='test@example.com')

def upgrade_schema(db=db):
    raise Exception("upgrade_schema() not implemented")

# TODO: Fix this so we can destroy older schemas that contain extraneous tables
#       without problems; right now, they would be left dangling unless we
#       upgraded first :o
def destroy_schema(db=db):
    metadata = get_metadata(db=db)
    metadata.drop_all(db)




import textwrap, os

def generate_mod_wsgi_config():
    user = 'git' # TODO: conf value?
    group = 'git' # TODO: conf value?
    venv = os.getenv('VIRTUAL_ENV', None)
    wsgi_dir = os.path.dirname(__file__)
    wsgi = os.path.join(wsgi_dir, 'gitzebo.wsgi')

    output = []
    output += [textwrap.dedent("""
        # Write to a usable prefix...
        WSGISocketPrefix /var/run/wsgi
        """)]
    if venv:
        output += [textwrap.dedent("""
            # Set up virtualenv
            WSGIPythonHome {venv}
            """).format(venv=venv)]
    output += [textwrap.dedent("""
        # Make sure authentication information is passed
        # request.auth is None otherwise.
        WSGIPassAuthorization On
        
        # Set permissions for gitzebo dir
        <Directory {wsgi_dir}>
            Order allow,deny
            Allow from all
        </Directory>
        
        # Set up WSGI Alias
        WSGIDaemonProcess gitzebo user={user} group={group}
        WSGIProcessGroup gitzebo
        WSGIScriptAlias / {wsgi}
        """).format(
            wsgi=wsgi,
            wsgi_dir=wsgi_dir,
            user=user,
            group=group
        )]

    return "\n".join(output)



#!/usr/bin/env python
import os
from setuptools import setup
from glob import glob

# Utility function so that we can share documentation with Sphinx politely.
def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()
def requirements():
    """
    Halfheartedly parses ./requirements.txt for dependency information.
    """
    # Read requirements spec
    requirements = [x.strip() for x in read('requirements.txt').split("\n")]
    # Remove blank and comment lines
    requirements = [x for x in requirements if len(x) > 0 and x[0] != '#']
    # Return
    return requirements
def scripts():
    """
    Reads all *.py filenames in ./bin for use as setup(scripts=) argument.
    """
    setup_dir = os.path.dirname(__file__)
    script_dir = os.path.join(setup_dir, 'bin')
    scripts = glob(os.path.join(script_dir,'*'))
    scripts = [os.path.relpath(p, setup_dir) for p in scripts]
    scripts = [s for s in scripts if s not in ['.', '..']]
    return scripts

setup(
    name = "gitzebo",
    # TODO: How do we make this and version.py play nicely together without
    #       a MAJOR hack that winds up breaking automated build frameworks?
    #       (For example, we don't want to break fpm compatibility.)
    version="0.0.6",
    author="John Gilik",
    author_email="john@jgilik.com",
    description=("A minimal git management web application."),
    license="2-Clause BSD License",
    keywords="git",
    url="http://jgilik.com/gitzebo/",
    packages=['gitzebo'],
    include_package_data=True, # include templates, static, etc
    long_description=read('README.rst'),
    scripts=scripts(),
    install_requires=requirements(),
    # TODO: Probably should add some classifiers.
    classifiers=[
        "Development Status :: 3 - Alpha",
    ],
)




#!/usr/bin/env python
# We'll need the schema module to play with the database.
from schema import db
import schema
users = schema.get_table('users')
repo_acls = schema.get_table('repository_acls')

# We'll also need some SQLAlchemy functionality to write queries.
from sqlalchemy.sql import select

# We need hashlib to play with password hashes.
import hashlib

# random and string for salt generation
import string, random

def generate_salt(size=16, chars=string.ascii_letters + string.digits):
    """
    Generates a salt according to `this StackOverflow answer
    <http://stackoverflow.com/a/2257449>`_.
    """
    return ''.join(random.choice(chars) for x in range(size))

def hash_password(password, salt):
    """
    Salts and hashes the password, returning hex-encoded string of the
    SHA256 hash.
    """
    return hashlib.sha256(salt + password).hexdigest()

def create_user(name, password, commit_name=None, commit_email=None, can_create_users=False, can_create_repositories=False):
    salt = generate_salt()
    hash = hash_password(password, salt)
    if commit_name is None:
        commit_name = name
    if commit_email is None:
        commit_email = ''
    stmt = users.insert().values(
        user_name=name,
        commit_name=commit_name,
        commit_email=commit_email,
        pass_salt=salt,
        pass_hash=hash,
        can_create_users=can_create_users,
        can_create_repositories=can_create_repositories,
    )
    result = db.execute(stmt)
    return result.inserted_primary_key[0]

def delete_user(user_id):
    s = repo_acls.delete().where(repo_acls.c.user_id == user_id)
    result = db.execute(s)
    s = users.delete().where(users.c.user_id == user_id)
    result = db.execute(s)
    # TODO: should do some sort of return here...

def get_user(name):
    s = select([users]).where(users.c.user_name == name)
    result = db.execute(s)
    return result.fetchone()

def get_user_by_id(id):
    s = select([users]).where(users.c.user_id == id)
    result = db.execute(s)
    return result.fetchone()

def get_users():
    s = select([users])
    result = db.execute(s)
    return result.fetchall()

def verify_user(name, password):
    try:
        user = get_user(name)
    except KeyError:
        return None
    salt = user[users.c.pass_salt]
    correct_hash = user[users.c.pass_hash]
    attempt_hash = hash_password(password, salt)
    if correct_hash != attempt_hash:
        return None
    return user

